from typing import Any

from typing_extensions import Self

class _BaseStorage:
    def __init__(self) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __copy__(self) -> Self: ...
    def __deepcopy__(self, memo: Any) -> Self: ...

class int64(_BaseStorage): ...
class double(_BaseStorage): ...
class atomic_int64(_BaseStorage): ...
class unlimited(_BaseStorage): ...
class weight(_BaseStorage): ...
class mean(_BaseStorage): ...
class weighted_mean(_BaseStorage): ...
